from dkulib.parallelizer import DataFrameParallelizer
from dkulib.parallelizer.parallelizer import ErrorHandling
from requests.exceptions import HTTPError
import requests
import os
import re
import logging
from shapely.geometry import shape, MultiPolygon
from dkuconstants import N_WORKERS
from dkuconstants import TRANSPORT_MODE
from dkuconstants import KNOWN_ERROR_MESSAGES
import json

logger = logging.getLogger(__name__)


class GeoRouter:
    def __init__(self, api_endpoint_url, dku_license_id=None):
        self.query_method = "GET"
        self.api_host = api_endpoint_url
        self.dku_license_id = dku_license_id
        self.point_pattern = re.compile(r"POINT\s*\((.*) (.*)\)")

    def parse_geopoint(self, geop):
        return self.point_pattern.findall(geop)[0]

    def geojson_to_wkt(self, gjson):
        if isinstance(gjson, list):
            features = []
            for feature in gjson:
                features.append(shape(feature["geometry"]))
            geometry_collection = MultiPolygon(features)
        else:
            if gjson["coordinates"] == []:
                return ""
            geometry_collection = shape(gjson)
        return geometry_collection.wkt

    def decode_encoded_polyline(self, encoded):
        inv = 1.0 / 1e6
        decoded = []
        previous = [0, 0]
        i = 0
        # for each byte
        while i < len(encoded):
            # for each coord (lat, lon)
            ll = [0, 0]
            for j in [0, 1]:
                shift = 0
                byte = 0x20
                # keep decoding bytes until you have this coord
                while byte >= 0x20:
                    byte = ord(encoded[i]) - 63
                    i += 1
                    ll[j] |= (byte & 0x1f) << shift
                    shift += 5
                # get the final value adding the previous offset and remember it for the next
                ll[j] = previous[j] + (~(ll[j] >> 1) if ll[j] & 1 else (ll[j] >> 1))
                previous[j] = ll[j]
            # scale by the precision and chop off long coords also flip the positions so
            # its the far more standard lon,lat instead of lat,lon
            decoded.append([float('%.6f' % (ll[1] * inv * 10)), float('%.6f' % (ll[0] * inv * 10))])
        # hand back the list of coordinates
        return decoded

    def parse_route_response(self, response, distance_unit, get_itinerary):
        route_response = {
            "time": round(self.get_time_from_response(response), 3),
            "distance": round(self.get_distance_from_response(response, distance_unit), 3)
        }
        if get_itinerary is True:
            route_response["itinerary"] = self.geojson_to_wkt(self.get_itinerary_from_response(response))
        return route_response

    def parse_isochrone_response(self, response):
        return {
            "isochrone": self.geojson_to_wkt(self.get_isochrone_geometry_from_response(response)),
        }

    def log_debug_messages(self, response, params, data, headers):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("state: " + str(response.status_code))
            logger.debug("params: " + str(params))
            logger.debug("data: " + str(data))
            logger.debug("headers: " + str(headers))
            logger.debug("query_method : " + str(self.query_method))
            logger.debug("response: " + str(response.json()))

    def enrich_headers(self, headers):
        headers["X-DKU-LicenseId"] = self.dku_license_id

    def make_get_request(self, url, params, data, headers):
        self.enrich_headers(headers)
        response = requests.get(url=url, params=params, headers=headers)
        self.log_debug_messages(response, params, data, headers)
        try:
            response.raise_for_status()
        except HTTPError as err:
            raise_error_400(response)
            if response.content:
                err.args = (f"{err.args[0]}\nResponse content: {response.content}",)
            raise err
        return response

    def get_isochrone(self, coords, time_threshold, transport_mode):
        params = self.get_isochrone_querystring(coords, time_threshold, transport_mode)
        data = self.get_isochrone_data(coords, time_threshold, transport_mode)
        headers = {}

        response = self.make_get_request(url=self.get_isochrone_url(), params=params, data=data, headers=headers)
        self.raise_for_specific_isochrone(response)
        return self.parse_isochrone_response(response.json())

    def get_route(self, from_coords, to_coords, transport_mode, distance_unit, get_itinerary):
        params = self.get_route_querystring(from_coords, to_coords, transport_mode, get_itinerary)
        data = self.get_route_data(from_coords, to_coords, transport_mode)
        headers = {}

        response = self.make_get_request(url=self.get_route_url(), params=params, data=data, headers=headers)
        self.raise_for_specific_route(response)
        return self.parse_route_response(response.json(), distance_unit, get_itinerary)

    def response_for_empty_point(self, get_itinerary):
        route_response = {"time": None, "distance": None}
        if get_itinerary is True:
            route_response["itinerary"] = None
        return json.dumps(route_response)

    def create_parallelizer(self, func):
        return DataFrameParallelizer(
            function=func,
            exceptions_to_catch=(Exception,),
            verbose=True,
            error_handling=ErrorHandling.LOG,
            parallel_workers=N_WORKERS,
        )

    def process_df_for_routes(self, input_dataset, from_col, to_col, transport_mode, distance_unit, output_dataset,
                              stream_chunk_size, get_itinerary=False, transport_mode_col=None, postprocess_func=None, postprocess_func_args=None):
        def process_row_for_routes(row):
            if from_col == to_col:
                logging.info("Starting point and destination are equal")
            if transport_mode == TRANSPORT_MODE.PARSE_COL:
                try:
                    real_transport_mode = TRANSPORT_MODE(row[transport_mode_col])
                except:
                    real_transport_mode = row[transport_mode_col]
            else:
                real_transport_mode = transport_mode
            dep_geopoint = row[from_col]
            arr_geopoint = row[to_col]
            try:
                if str(dep_geopoint) == 'nan':
                    return self.response_for_empty_point(get_itinerary)
                else:
                    parsed_from_coords = self.parse_geopoint(dep_geopoint)
            except:
                raise SyntaxError("Parsing error: Input column '{0}' should contain geometries in WKT format.".format(from_col))
            try:
                if str(arr_geopoint) == 'nan':
                    return self.response_for_empty_point(get_itinerary)
                else:
                    parsed_to_coords = self.parse_geopoint(arr_geopoint)
            except:
                raise SyntaxError("Parsing error: Input column '{0}' should contain geometries in WKT format.".format(to_col))
            return json.dumps(
                self.get_route(
                    from_coords=parsed_from_coords,
                    to_coords=parsed_to_coords,
                    transport_mode=real_transport_mode,
                    distance_unit=distance_unit,
                    get_itinerary=get_itinerary
                )
            )

        router_parallelizer = self.create_parallelizer(process_row_for_routes)
        return router_parallelizer.run_streamed(input_dataset, output_dataset, stream_chunk_size=stream_chunk_size, postprocess_func=postprocess_func, postprocess_func_args=postprocess_func_args)

    def process_df_for_isochrones(self, input_dataset, output_dataset, stream_chunk_size, coords_col, time_threshold, transport_mode, transport_mode_col=None,
                                  postprocess_func=None, postprocess_func_args=None):
        def process_row_for_isochrone(row):
            if transport_mode == TRANSPORT_MODE.PARSE_COL:
                try:
                    real_transport_mode = TRANSPORT_MODE(row[transport_mode_col])
                except:
                    real_transport_mode = row[transport_mode_col]
            else:
                real_transport_mode = transport_mode
            geopoint = row[coords_col]
            try:
                if str(geopoint) == 'nan':
                    return json.dumps({"isochrone": None})
                else:
                    parsed_geopoint = self.parse_geopoint(geopoint)
            except:
                raise SyntaxError("Parsing error: Input column '{0}' should contain geometries in WKT format.".format(coords_col))
            return json.dumps(
                self.get_isochrone(
                    coords=parsed_geopoint,
                    time_threshold=float(time_threshold),
                    transport_mode=real_transport_mode,
                )
            )
        router_parallelizer = self.create_parallelizer(process_row_for_isochrone)

        return router_parallelizer.run_streamed(input_dataset, output_dataset, stream_chunk_size=stream_chunk_size, postprocess_func=postprocess_func, postprocess_func_args=postprocess_func_args)

    def raise_for_specific_route(self, response):
        return

    def raise_for_specific_isochrone(self, response):
        return

    def get_route_data(self, from_coords, to_coords, global_transport_mod):
        return {}

    def get_route_querystring(self, from_coords, to_coords, global_transport_mod, get_itinerary):
        return {}

    def get_route_url(self):
        return os.path.join(self.api_host, self.route_endpoint)

    def get_isochrone_data(self, coords, time_threshold, transport_mode):
        return {}

    def get_isochrone_querystring(self, coords, time_threshold, transport_mode):
        return {}

    def get_isochrone_url(self):
        return os.path.join(self.api_host, self.isochrone_endpoint)

    def get_isochrone_geometry_from_response(self, response):
        raise NotImplementedError()

    def get_itinerary_from_response(self, response):
        raise NotImplementedError()

    def get_time_from_response(self, response):
        raise NotImplementedError()

    def get_distance_from_response(self, response, distance_unit):
        raise NotImplementedError()


def raise_error_400(response):
    json_response = {}
    try:
        if response.status_code == 400 and response.content:
            json_response = response.json()
    except Exception as err:
        logger.warning("Error parsing response:{}".format(err))
    raw_error_message = json_response.get("message")
    if raw_error_message:
        raise ValueError(improve_error_message(raw_error_message))


def improve_error_message(message):
    improved_message = message
    for message_sample in KNOWN_ERROR_MESSAGES:
        if message_sample in message:
            improved_message = "{} / {}".format(
                KNOWN_ERROR_MESSAGES.get(message_sample),
                message
            )
            return improved_message
    return improved_message
